/*
 * 08/06/2003
 *
 * image.cpp
 * Classe permettant de stocker une image
 *
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "image.h"

// ============
// Constructeur
// ============
CImage::CImage() {
	data = NULL;
	width = 0;
	height = 0;
	bitCount = 0;
}

// ============
// Constructeur
// ============
CImage::CImage(int w, int h, int d)
{
	int nbplan = d / 8;
	data = new unsigned char[w * h * nbplan];
	width = w;
	height = h;
	bitCount = d;
}

// ============
// Constructeur
// ============
CImage::CImage(unsigned char *da, int w, int h, int d)
{
	int nbplan = d / 8;
	data = new unsigned char[w * h * nbplan];
	memcpy((void*)data, (void*)da, w * h * nbplan * sizeof(unsigned char));
	width = w;
	height = h;
	bitCount = d;
}

// ===========
// Destructeur
// ===========
CImage::~CImage()
{
	if (data) delete data;
}

// =============================
// Retourne une copie de l'image
// =============================
CImage * CImage::Copy()
{
	return new CImage(data, width, height, bitCount);
}

// ================================================================
// Fonction permettant de tester si un entier est une puisance de 2
// ================================================================
bool CImage::CheckSize(int x)
{
    if (x == 2	 || x == 4 || 
        x == 8	 || x == 16 || 
        x == 32  || x == 64 ||
        x == 128 || x == 256 || 
	x == 512 || x == 1024)
        return true;
    else return false;
}

// ====================================================================
// Retourne le format OpenGL de l'image (GL_RGB, GL_RGBA, GL_LUMINANCE)
// ====================================================================
GLenum CImage::GetTextureFormat()
{
	GLenum texFormat;
	// Format de l'image RGB, RGBA ou niveau de gris
	switch (bitCount) {
		case 8: texFormat = GL_LUMINANCE; break;
		case 32: texFormat = GL_RGBA; break;
		default: texFormat = GL_RGB;
	}
	
	return texFormat;
}

// ==================================================
// Retourne le type de l'image en fonction de son nom
// ==================================================
int CImage::GetType(const char *name)
{
	int i;
	int res;
	for (i = 0 ; name[i] != '\0'; i++);

	while (name[i] != '.' || i==0)
	{
		i--;
	}
	
	if (i==0) return INVALID;

	char *ext;
	
#ifdef WIN32
	ext = strupr(strdup(name+i));
	if (strcmp(ext, ".BMP") == 0) res = BMP;
	else if (strcmp(ext, ".TGA") == 0) res = TGA;
	else if (strcmp(ext, ".PPM") == 0) res = PPM;
	else if (strcmp(ext, ".PGM") == 0) res = PGM;
	else res = INVALID;
#else
	ext = strdup(name+i);
	if (strcasecmp(ext, ".BMP") == 0) res = BMP;
	else if (strcasecmp(ext, ".TGA") == 0) res = TGA;
	else if (strcasecmp(ext, ".PPM") == 0) res = PPM;
	else if (strcasecmp(ext, ".PGM") == 0) res = PGM;
	else res = INVALID;
#endif

	delete ext;

	return res;
}

// ================================================
// Fonction permettant de lire un fichier image TGA
// 8, 16, 24 ou 32 bits
// Le format 32 bits est convertit en 24 bits
// ================================================
// Le format Targa TGA est divis en 2 partie l'en-tte et les donnes.
// L'en-tte contient 12 champs et est dfinie par la structure suivante :
//
// typedef struct tagTARGAFILEHEADER
// {
//	unsigned char imageIDLength;	// number of characters in identification field;
//									// 0 denotes no identification field is included
//	unsigned char colorMapType;		// type of color map; always 0
//	unsigned char imageTypeCode;	// uncompressed RGB is 2;
//									// uncompressed grayscale is 3
//	short int colorMapOrigin;		// origin of color map (lo-hi); always 0
//	short int colorMapLength;		// length of color map (lo-hi); always 0
//	short int colorMapEntrySize;	// color map entry size (lo-hi); always 0
//	short int imageXOrigin;			// x coordinate of lower-left corner of image
//									// (lo-hi); always 0
//	short int imageYOrigin;			// y coordinate of lower-left corner of image
//									// (lo-hi); always 0
//	short int imageWidth;			// width of image in pixels (lo-hi)
//	short int imageHeight;			// height of image in pixels (lo-hi)
//	unsigned char bitCount;			// number of bits; 16, 24, 32
//	unsigned char imageDescriptor;	// 24 bit = 0x00; 32 bit = 0x08
// } TARGAFILEHEADER
int CImage::LoadTGA(CImage *img, const char *filename)
{
	FILE *filePtr;				// pointeur sur le fichier
	unsigned char ucharBad;		// poubelle pour les unsigned char
	short int sintBad;			// poubelle pour les short int
	int colorMode;				// 4 pour RGBA ou 3 pour RGB
	long imageSize;				// taille de l'image TGA
	long imageIdx;				// compteur
	unsigned char colorSwap;	// variable d'change
	unsigned short pixels = 0;	// un pixel d'une image 16 bits
	int r, g, b;				// composantes RGB
	
	// ouverture du fichier TGA
	if ((filePtr = fopen (filename, "rb")) == NULL) {
        return IMG_FILE_NOT_FOUND;
	}
	
	// lecture des 2 premiers champs dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// lecture du type de l'image
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	if ((ucharBad != 2) && (ucharBad != 3)) {
		fclose(filePtr);
		return IMG_BAD_IMAGE_TYPE;
	}
	// lecture des 5 champs suivants dont on n'a pas besoin
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	
	// lecture des dimensions de l'image
	fread(&(img->width), sizeof(short int), 1, filePtr);
	fread(&(img->height), sizeof(short int), 1, filePtr);
	fread(&(img->bitCount), sizeof(unsigned char), 1, filePtr);
	
	// lecture du dernier champ dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// on ne lit que les images 8, 16, 24 ou 32 bits
	if (img->bitCount != 32 && img->bitCount != 24 && img->bitCount != 8 && img->bitCount != 16) {
		fclose(filePtr);
		return IMG_BAD_BITS;
	}
	
	// image 8, 24 ou 32 bits
	if(img->bitCount == 24 || img->bitCount == 8 || img->bitCount == 32)
	{
		// colorMode -> 3 = BGR, 4 = BGRA
		colorMode = img->bitCount / 8;
		imageSize = img->width * img->height * colorMode;
	
		// allocation de la mmoire pour les donnes de l'image
		//m_data = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
		img->data = new unsigned char[imageSize];

		// lecture des donnes de l'image
		fread(img->data, sizeof(unsigned char), imageSize, filePtr);
		if (img->data == NULL) {
        		fclose(filePtr);
			return IMG_BAD_DATA;
		}
	
		// dans les images TGA les couleurs des pixels sont en mode BGR
		// transformation en mode RGB
		if (img->bitCount != 8)
		for (imageIdx=0; imageIdx < imageSize; imageIdx += colorMode)
		{
			colorSwap = img->data[imageIdx];
			img->data[imageIdx] = img->data[imageIdx + 2];
			img->data[imageIdx + 2] = colorSwap;
		}
	}
	else { // image 16 bits

		img->bitCount = 24;
		colorMode = 3;	// mode RGB
		imageSize = img->width * img->height * colorMode;
		
		// allocation de la mmoire pour les donnes de l'image
		//tgaFile->imageData = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
		img->data = new unsigned char[imageSize];
		
		// en 16 bits chaque pixel est contenu dans un "unsigned short"
		// boucle sur tous les pixels de l'image
		for (imageIdx = 0; imageIdx < img->width * img->height; imageIdx++)
		{
			// lecture du pixel courant
			fread(&pixels, sizeof(unsigned short), 1, filePtr);
				
			// pour convertir 16 bits en composantes RGB
			// chaque composante occupe 5 bits
			// ex: pixels = 0x61E8 = 0110000111101000
			// on utilise un masque 0x1F = 11111
			// 0110000111101000 & 11111 = 01000
			// on effectue un dcallage de 3 bits pour avoir un nombre sur 8 bits
			// 01000 << 3 = 01000000 = 0x40 = 64 --> composante bleu
			// on dcale ensuite de 5 bits pour arriver  la prochaine
			// composante et on ritre la mme opration
			// 0110000111101000 >> 5 = 0000001100001111
			// 0000001100001111 & 11111 = 01111
			// 01111 << 3 = 01111000 = 0x78 = 120 --> composante vert
			// ...etc...
			b = (pixels & 0x1f) << 3;
			g = ((pixels >> 5) & 0x1f) << 3;
			r = ((pixels >> 10) & 0x1f) << 3;
				
			// on assigne ensuite les composantes au tableau de donnes
			img->data[imageIdx * 3 + 0] = r;
			img->data[imageIdx * 3 + 1] = g;
			img->data[imageIdx * 3 + 2] = b;
		}
	}

	// on convertit les images 32 bits en 24 bits
	if(img->bitCount == 32)
	{
		colorMode = img->bitCount / 8;
		imageSize = img->width * img->height * colorMode;

		unsigned char * temp = new unsigned char[img->width * img->height * 3];

		int i = 0;
		for (imageIdx=0; imageIdx < imageSize; imageIdx += colorMode)
		{
			temp[i] = img->data[imageIdx];
			temp[i + 1] = img->data[imageIdx + 1];
			temp[i + 2] = img->data[imageIdx + 2];
			i+=3;
		}

		delete img->data;

		img->data = temp;

		img->bitCount = 24;
		colorMode = 3;
		imageSize = img->width * img->height * colorMode;
	}

	// Symtrie d'axe verticale
	CImage *temp = new CImage(img->width, img->height, img->bitCount);
	int x,y;
	int width = img->width*colorMode;
	for (long i = 0; i < imageSize; i+=colorMode) {
		x = i % width;
		y = i / width;
		temp->data[(img->height-1-y)*width + x] = img->data[i];
		if (colorMode == 3) {
			temp->data[(img->height-1-y)*width + x +1] = img->data[i+1];
			temp->data[(img->height-1-y)*width + x +2] = img->data[i+2];
		}
	}

	delete img;
	img = temp->Copy();
	delete temp;

	// fermeture du fichier
	fclose(filePtr);
	
	return IMG_OK;
}

// ============================================================
// Fonction enregistrant une image TGA dans le fichier filename
// ============================================================
int CImage::SaveTGA(CImage *img, const char *filename)
{
	unsigned char byteSkip;		// utilis pour mettre  0 certains champs
	short int shortSkip;		// utilis pour mettre  0 certains champs
	unsigned char imageType;	// type de l'image
	int colorMode;				// color mode
	unsigned char colorSwap;	// utilis pour la conversion RGB - BGR
	int imageIdx;				// compteur
	unsigned char bitDepth;		// image 24 bits
	long imageSize;				// taille de l'image
	FILE *filePtr;				// pointeur sur le fichier
	
	// cration du fichier en mode criture binaire
	if (!(filePtr = fopen (filename, "wb"))) {
		return IMG_FILE_NOT_CREATE;
	}
	
	imageType = 2;	// RGB uncompressed
	bitDepth = img->bitCount;	// bitdepth
	colorMode = img->bitCount / 8;	// color mode
	
	byteSkip = 0;
	shortSkip = 0;
	
	// criture des 2 premiers champs  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// criture du type de l'image
	fwrite(&imageType, sizeof(unsigned char), 1, filePtr);
	
	// criture des 5 champs suivants  0
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	
	// criture des dimensions de l'image
	fwrite(&img->width, sizeof(short int), 1, filePtr);
	fwrite(&img->height, sizeof(short int), 1, filePtr);
	fwrite(&bitDepth, sizeof(unsigned char), 1, filePtr);
	
	// criture du dernier champ  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// calcul de la taille de l'image
	imageSize = img->width * img->height * colorMode;
	
	CImage *temp = img->Copy();

	// conversion des donnes du mode RGB vers le mode BGR
	if (colorMode == 3)
	for (imageIdx = 0; imageIdx < imageSize ; imageIdx += colorMode)
	{
		colorSwap = temp->data[imageIdx];
		temp->data[imageIdx] = temp->data[imageIdx + 2];
		temp->data[imageIdx + 2] = colorSwap;
	}
	
	// Symtrie d'axe verticale
	int x,y;
	int width = img->width*colorMode;
	for (long i = 0; i < imageSize; i+=colorMode) {
		x = i % width;
		y = i / width;
		temp->data[(img->height-1-y)*width + x] = img->data[i];
		if (colorMode == 3) {
			temp->data[(img->height-1-y)*width + x +1] = img->data[i+1];
			temp->data[(img->height-1-y)*width + x +2] = img->data[i+2];
		}
	}

	// criture des donnes de l'image dans le fichier
	fwrite(temp->data, sizeof(unsigned char), imageSize, filePtr);
	
	delete temp;

	// fermeture du fichier
	fclose(filePtr);
	
	return IMG_OK;	
}

// ============================================================
// Fonction enregistrant une image BMP dans le fichier filename
// ============================================================
int CImage::SaveBMP(CImage *img, const char *filename)
{
	FILE			 *filePtr;			// file pointer
	BITMAPFILEHEADER bitmapFileHeader;	// bitmap file header
	BITMAPINFOHEADER bitmapInfoHeader;	// bitmap info header
	int				 imageIdx;			// used for swapping RGB->BGR
	unsigned char colorSwap;	// utilis pour la conversion RGB - BGR

	// cration du fichier en mode criture binaire
	filePtr = fopen(filename, "wb");
	if (!filePtr)
		return IMG_FILE_NOT_CREATE;

	// bitmap file header
	bitmapFileHeader.bfSize = sizeof(BITMAPFILEHEADER);
	bitmapFileHeader.bfType = 0x4D42;
	bitmapFileHeader.bfReserved1 = 0;
	bitmapFileHeader.bfReserved2 = 0;
	bitmapFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (img->bitCount == 8) bitmapFileHeader.bfOffBits += 256*sizeof(RGBQUAD);

	// bitmap information header
	bitmapInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
	bitmapInfoHeader.biPlanes = 1;
	bitmapInfoHeader.biBitCount = img->bitCount;					
	bitmapInfoHeader.biCompression = BI_RGB;				// no compression
	bitmapInfoHeader.biSizeImage = img->width * abs(img->height) * (img->bitCount/8);
	bitmapInfoHeader.biXPelsPerMeter = 0;
	bitmapInfoHeader.biYPelsPerMeter = 0;
	bitmapInfoHeader.biClrUsed = 0;
	if (img->bitCount == 8) bitmapInfoHeader.biClrUsed = 256;
	bitmapInfoHeader.biClrImportant = 0;
	bitmapInfoHeader.biWidth = img->width;						// bitmap width
	bitmapInfoHeader.biHeight = img->height;					// bitmap height

	CImage *temp = img->Copy();
	int colorMode = img->bitCount/8;

	// Symtrie d'axe verticale
	int x,y;
	int width = img->width*colorMode;
	for (long i = 0; i < bitmapInfoHeader.biSizeImage; i+=colorMode) {
		x = i % width;
		y = i / width;
		temp->data[(img->height-1-y)*width + x] = img->data[i];
		if (colorMode == 3) {
			temp->data[(img->height-1-y)*width + x +1] = img->data[i+1];
			temp->data[(img->height-1-y)*width + x +2] = img->data[i+2];
		}
	}

	// conversion des donnes du mode RGB vers le mode BGR
	if (colorMode == 3)
	for (imageIdx = 0; imageIdx < bitmapInfoHeader.biSizeImage ; imageIdx += colorMode)
	{
		colorSwap = temp->data[imageIdx];
		temp->data[imageIdx] = temp->data[imageIdx + 2];
		temp->data[imageIdx + 2] = colorSwap;
	}

	// write the bitmap file header
	fwrite(&bitmapFileHeader, 1, sizeof(BITMAPFILEHEADER), filePtr);

	// write the bitmap info header
	fwrite(&bitmapInfoHeader, 1, sizeof(BITMAPINFOHEADER), filePtr);

	// Pour les images en niveau de gris il faut ecrire
	// un tableau contenant les diffrents niveaux de gris
	if (colorMode == 1) {
		RGBQUAD colors[256];	
		for (int i=0; i<256; i++) {
			colors[i].rgbRed = i;
			colors[i].rgbGreen = i;
			colors[i].rgbBlue = i;
		}
		fwrite(&colors, 1, 256*sizeof(RGBQUAD), filePtr);
	}

	// write the image data
	fwrite(temp->data, 1, bitmapInfoHeader.biSizeImage, filePtr);

	delete temp;

	// close our file
	fclose(filePtr);

	return IMG_OK;
}

